@@ -21,6 +21,7 @@ from concurrent.futures import Future
 from gremlin_python.driver import client, serializer
 from gremlin_python.driver.remote_connection import (
     RemoteConnection, RemoteTraversal, RemoteTraversalSideEffects)
+from gremlin_python.process.strategies import OptionsStrategy
 
 __author__ = 'David M. Brown (davebshow@gmail.com)'
 
@@ -31,7 +32,7 @@ class DriverRemoteConnection(RemoteConnection):
                  transport_factory=None, pool_size=None, max_workers=None,
                  username="", password="", message_serializer=None,
                  graphson_reader=None, graphson_writer=None,
-                 headers=None):
+                 headers=None, max_content_length=None):
         if message_serializer is None:
             message_serializer = serializer.GraphSONMessageSerializer(
                 reader=graphson_reader,
@@ -44,7 +45,8 @@ class DriverRemoteConnection(RemoteConnection):
                                      message_serializer=message_serializer,
                                      username=username,
                                      password=password,
-                                     headers=headers)
+                                     headers=headers,
+                                     max_content_length=max_content_length)
         self._url = self._client._url
         self._traversal_source = self._client._traversal_source
 
@@ -52,25 +54,36 @@ class DriverRemoteConnection(RemoteConnection):
         self._client.close()
 
     def submit(self, bytecode):
-        result_set = self._client.submit(bytecode)
+        result_set = self._client.submit(bytecode, request_options=self._extract_request_options(bytecode))
         results = result_set.all().result()
-        side_effects = RemoteTraversalSideEffects(result_set.request_id,
-                                                  self._client)
+        side_effects = RemoteTraversalSideEffects(result_set.request_id, self._client,
+                                                  result_set.status_attributes)
         return RemoteTraversal(iter(results), side_effects)
 
     def submitAsync(self, bytecode):
         future = Future()
-        future_result_set = self._client.submitAsync(bytecode)
+        future_result_set = self._client.submitAsync(bytecode, request_options=self._extract_request_options(bytecode))
 
         def cb(f):
             try:
                 result_set = f.result()
                 results = result_set.all().result()
-                side_effects = RemoteTraversalSideEffects(result_set.request_id,
-                                                          self._client)
+                side_effects = RemoteTraversalSideEffects(result_set.request_id, self._client,
+                                                          result_set.status_attributes)
                 future.set_result(RemoteTraversal(iter(results), side_effects))
             except Exception as e:
                 future.set_exception(e)
 
         future_result_set.add_done_callback(cb)
         return future
+
+    @staticmethod
+    def _extract_request_options(bytecode):
+        options_strategy = next((x for x in bytecode.source_instructions
+                                 if x[0] == "withStrategies" and type(x[1]) is OptionsStrategy), None)
+        request_options = None
+        if options_strategy:
+            allowed_keys = ['evaluationTimeout', 'scriptEvaluationTimeout', 'batchSize', 'requestId', 'userAgent']
+            request_options = {allowed: options_strategy[1].configuration[allowed] for allowed in allowed_keys
+                               if allowed in options_strategy[1].configuration}
+        return request_options
