@@ -1,26 +1,27 @@
-'''
-Licensed to the Apache Software Foundation (ASF) under one
-or more contributor license agreements.  See the NOTICE file
-distributed with this work for additional information
-regarding copyright ownership.  The ASF licenses this file
-to you under the Apache License, Version 2.0 (the
-"License"); you may not use this file except in compliance
-with the License.  You may obtain a copy of the License at
-
-http://www.apache.org/licenses/LICENSE-2.0
-
-Unless required by applicable law or agreed to in writing,
-software distributed under the License is distributed on an
-"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
-KIND, either express or implied.  See the License for the
-specific language governing permissions and limitations
-under the License.
-'''
+#
+# Licensed to the Apache Software Foundation (ASF) under one
+# or more contributor license agreements.  See the NOTICE file
+# distributed with this work for additional information
+# regarding copyright ownership.  The ASF licenses this file
+# to you under the Apache License, Version 2.0 (the
+# "License"); you may not use this file except in compliance
+# with the License.  You may obtain a copy of the License at
+#
+# http://www.apache.org/licenses/LICENSE-2.0
+#
+# Unless required by applicable law or agreed to in writing,
+# software distributed under the License is distributed on an
+# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
+# KIND, either express or implied.  See the License for the
+# specific language governing permissions and limitations
+# under the License.
+#
 
 import sys
+import copy
 from .traversal import Traversal
 from .traversal import TraversalStrategies
-from .strategies import VertexProgramStrategy
+from .strategies import VertexProgramStrategy, OptionsStrategy
 from .traversal import Bytecode
 from ..driver.remote_connection import RemoteStrategy
 from .. import statics
@@ -70,16 +71,25 @@ class GraphTraversalSource(object):
         source.bytecode.add_source("withStrategies", *args)
         return source
 
-    def with_(self, *args):
-        source = self.get_graph_traversal_source()
-        source.bytecode.add_source("with", *args)
-        return source
-
     def withoutStrategies(self, *args):
         source = self.get_graph_traversal_source()
         source.bytecode.add_source("withoutStrategies", *args)
         return source
 
+    def with_(self, k, v=None):
+        source = self.get_graph_traversal_source()
+        options_strategy = next((x for x in source.bytecode.source_instructions
+                                if x[0] == "withStrategies" and type(x[1]) is OptionsStrategy), None)
+
+        val = True if v is None else v
+        if options_strategy is None:
+            options_strategy = OptionsStrategy({k: val})
+            source = self.withStrategies(options_strategy)
+        else:
+            options_strategy[1].configuration[k] = val
+
+        return source
+
     def withRemote(self, remote_connection):
         source = self.get_graph_traversal_source()
         source.traversal_strategies.add_strategies([RemoteStrategy(remote_connection)])
@@ -139,8 +149,14 @@ class GraphTraversal(Traversal):
             raise TypeError("Index must be int or slice")
 
     def __getattr__(self, key):
+        if key.startswith('__'):
+            raise AttributeError(
+                'Python magic methods or keys starting with double underscore cannot be used for Gremlin sugar - prefer values(' + key + ')')
         return self.values(key)
 
+    def clone(self):
+        return GraphTraversal(self.graph, self.traversal_strategies, copy.deepcopy(self.bytecode))
+
     def V(self, *args):
         self.bytecode.add_step("V", *args)
         return self
@@ -229,6 +245,10 @@ class GraphTraversal(Traversal):
         self.bytecode.add_step("drop", *args)
         return self
 
+    def elementMap(self, *args):
+        self.bytecode.add_step("elementMap", *args)
+        return self
+
     def emit(self, *args):
         self.bytecode.add_step("emit", *args)
         return self
@@ -357,6 +377,10 @@ class GraphTraversal(Traversal):
         self.bytecode.add_step("min", *args)
         return self
 
+    def none(self, *args):
+        self.bytecode.add_step("none", *args)
+        return self
+
     def not_(self, *args):
         self.bytecode.add_step("not", *args)
         return self
@@ -571,8 +595,16 @@ class GraphTraversal(Traversal):
         self.bytecode.add_step("sum", *args)
         return self
 
+class MagicType(type):
+
+    def __getattr__(cls, k):
+        if k.startswith('__'):
+            raise AttributeError(
+                'Python magic methods or keys starting with double underscore cannot be used for Gremlin sugar - prefer values(' + k + ')')
+        return __.values(k)
 
 class __(object):
+    __metaclass__ = MagicType
     graph_traversal = GraphTraversal
 
     @classmethod
@@ -663,6 +695,10 @@ class __(object):
     def drop(cls, *args):
         return cls.graph_traversal(None, None, Bytecode()).drop(*args)
 
+    @classmethod
+    def elementMap(cls, *args):
+        return cls.graph_traversal(None, None, Bytecode()).elementMap(*args)
+
     @classmethod
     def emit(cls, *args):
         return cls.graph_traversal(None, None, Bytecode()).emit(*args)
@@ -1046,6 +1082,10 @@ def drop(*args):
     return __.drop(*args)
 
 
+def elementMap(*args):
+    return __.elementMap(*args)
+
+
 def emit(*args):
     return __.emit(*args)
 
@@ -1388,6 +1428,8 @@ statics.add_static('dedup', dedup)
 
 statics.add_static('drop', drop)
 
+statics.add_static('elementMap', elementMap)
+
 statics.add_static('emit', emit)
 
 statics.add_static('filter_', filter_)
